from pydrake.trajectories import CompositeTrajectory, Trajectory

from airo_drake import shift_drake_trajectory_in_time


def concatenate_drake_trajectories(trajectories: list[Trajectory]) -> CompositeTrajectory:
    """Concatenate multiple trajectories in time. The start time will be the start time of the first trajectory.

    This is achieved by shifting the times of the trajectories to make them continuous.

    Args:
        trajectories: A list of Drake trajectories to be concatenated.

    Returns:
        A CompositeTrajectory representing the concatenated sequence.
    """

    if not trajectories:
        raise ValueError("Cannot concatenate an empty list of trajectories.")

    offset = 0.0
    shifted_trajectories = []
    for trajectory in trajectories:
        shifted_trajectories.append(shift_drake_trajectory_in_time(trajectory, offset))
        offset += trajectory.end_time() - trajectory.start_time()

    return CompositeTrajectory(shifted_trajectories)  # type: ignore



import numpy as np
from airo_typing import (
    InverseKinematicsFunctionType,
    JointConfigurationCheckerType,
    JointConfigurationType,
    JointPathType,
    PosePathType,
)
from loguru import logger

from airo_drake import joint_path_has_large_jumps
from airo_drake.path.analysis import find_closest_configuration


def calculate_path_ik_solutions(
    tcp_path: PosePathType, inverse_kinematics_fn: InverseKinematicsFunctionType
) -> list[list[JointConfigurationType]]:
    path_ik_solutions = [inverse_kinematics_fn(tcp_pose) for tcp_pose in tcp_path]
    return path_ik_solutions


def create_paths_from_closest_solutions(
    path_joint_solutions: list[list[JointConfigurationType]],
) -> list[JointPathType]:
    """Constructs paths by iteratively connecting closest joint configurations."""

    paths = []
    start_configurations = path_joint_solutions[0]
    for start_configuration in start_configurations:
        path = [start_configuration]
        for joint_solutions in path_joint_solutions[1:]:
            # logger.info(f"len(joint_solutions): {len(joint_solutions)}")
            if len(joint_solutions) == 0:
                logger.warning("Could not create paths, one of the states has no joint solutions.")
                return []
            closest_config = find_closest_configuration(path[-1], joint_solutions)
            path.append(closest_config)
        paths.append(np.array(path))
    return paths


def calculate_joint_paths(
    tcp_path: PosePathType, inverse_kinematics_fn: InverseKinematicsFunctionType
) -> list[JointPathType]:
    path_ik_solutions = calculate_path_ik_solutions(tcp_path, inverse_kinematics_fn)
    joint_path_candidates = create_paths_from_closest_solutions(path_ik_solutions)
    joint_paths = [path for path in joint_path_candidates if not joint_path_has_large_jumps(path)]
    return joint_paths


def calculate_valid_joint_paths(
    tcp_path: PosePathType,
    inverse_kinematics_fn: InverseKinematicsFunctionType,
    is_state_valid_fn: JointConfigurationCheckerType,
) -> list[JointPathType]:

    joint_paths = calculate_joint_paths(tcp_path, inverse_kinematics_fn)

    # Only keep paths where all joint configurations are valid
    valid_joint_paths = []

    for joint_path in joint_paths:
        if all(is_state_valid_fn(joint_configuration) for joint_configuration in joint_path):
            valid_joint_paths.append(joint_path)

    return valid_joint_paths



import numpy as np
from airo_typing import RotationMatrixType, Vector3DType
from pydrake.geometry import Cylinder, Meshcat, MeshcatCone, Rgba
from pydrake.math import RigidTransform, RollPitchYaw


def orientation_with_Z_axis(Z: Vector3DType) -> RotationMatrixType:
    Z = np.array(Z)
    Z = Z / np.linalg.norm(Z)  # normalize just to be sure

    X = np.array([1, 0, 0])

    # handle case where direction is parallel to X
    if np.all(np.abs(Z - X) < 1e-6):
        X = np.array([0, 1, 0])

    # make X orthogonal to Z
    X = X - Z * np.dot(X, Z)
    X = X / np.linalg.norm(X)

    # make Y orthogonal to Z and X
    Y = np.cross(Z, X)

    return np.column_stack([X, Y, Z])


def visualize_arrow(
    meshcat: Meshcat,
    name: str,
    start: Vector3DType,
    direction: Vector3DType | None = None,
    end: Vector3DType | None = None,
    length: float = 1.0,
    radius: float = 0.005,
    color: Rgba | None = None,
) -> None:
    """Visualizes an arrow in meshcat.

    Args:
        meshcat: The MeshCat instance to add the visualization to.
        name: A name or path for the arrow in the MeshCat hierarchy, can be used to delete it later.
        start: The starting point of the arrow to visualization.
        direction: The direction of the arrow to visualize.
        end: The end point of the arrow to visualize, mutually exclusive with direction.
        length: Length of the arrow, not used if end is provided.
        radius: Radius of the cylinder.
        color: Color of the arrow.
    """
    start = np.array(start)

    if direction is None and end is None:
        raise ValueError("To visualize an arrow either direction or end must be provided.")

    if direction is not None and end is not None:
        raise ValueError("To visualize an arrow either direction or end must be provided, not both.")

    if direction is None:
        direction = np.array(end) - start
        length = float(np.linalg.norm(direction))

    direction = np.array(direction)

    cone_height = 4 * radius  # rule of thumb
    cone_width = cone_height / 2  # rule of thumb
    cylinder_height = length - cone_height

    orientation = orientation_with_Z_axis(direction)

    pose = np.identity(4)
    pose[:3, :3] = orientation
    pose[:3, 3] = start

    if color is None:
        color_from_direction = np.abs(direction)
        color_from_direction = color_from_direction / np.linalg.norm(color_from_direction)
        color_from_direction = color_from_direction / np.max(color_from_direction)
        color = Rgba(*color_from_direction, 1)  # type: ignore

    arrow_transform = RigidTransform(pose)
    meshcat.SetTransform(f"{name}/arrow/", arrow_transform)

    cylinder_transform = RigidTransform(p=[0, 0, cylinder_height / 2])  # type: ignore
    meshcat.SetObject(f"{name}/arrow/cylinder", Cylinder(radius, cylinder_height), color)
    meshcat.SetTransform(f"{name}/arrow/cylinder", cylinder_transform)

    cone_transform = RigidTransform(rpy=RollPitchYaw(np.deg2rad([180, 0, 0])), p=[0, 0, cylinder_height + cone_height])  # type: ignore
    meshcat.SetTransform(f"{name}/arrow/cone/", cone_transform)
    meshcat.SetObject(f"{name}/arrow/cone", MeshcatCone(cone_height, cone_width, cone_width), color)



import numpy as np
from airo_typing import DualArmTrajectory, JointPathType, SingleArmTrajectory
from loguru import logger
from pydrake.geometry import Meshcat
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.planning import RobotDiagram
from pydrake.systems.framework import Context
from pydrake.trajectories import PiecewisePolynomial, Trajectory

from airo_drake import joint_trajectory_to_drake


def animate_joint_trajectory(
    meshcat: Meshcat,
    robot_diagram: RobotDiagram,
    arm_index: ModelInstanceIndex,
    joint_trajectory: SingleArmTrajectory | Trajectory,
    context: Context | None = None,
) -> None:
    """Publish a recorded animation to meshcat where the robot arm follows the provided joint trajectory.

    Args:
        meshcat: The MeshCat instance to add the visualization to.
        robot_diagram: The robot diagram.
        arm_index: The index of the robot arm to animate.
        joint_trajectory: The joint trajectory to animate.
        context: The context to use for the animation. If None, a new default context will be used (with all other joint angles equal to 0).
    """
    if joint_trajectory is None:
        logger.warning("Asked to animate joint trajectory, but none provided.")

    if isinstance(joint_trajectory, SingleArmTrajectory):
        joint_trajectory = joint_trajectory_to_drake(joint_trajectory)

    assert isinstance(joint_trajectory, Trajectory)  # For mypy, but extra check we have a Drake Trajectory now

    if context is None:
        # Create a new context just for the recording
        context = robot_diagram.CreateDefaultContext()
    plant = robot_diagram.plant()
    plant_context = plant.GetMyContextFromRoot(context)

    fps = 60.0

    duration = joint_trajectory.end_time()

    # TODO can support trajectories that don't start at 0?
    if not np.isclose(joint_trajectory.start_time(), 0.0):
        logger.warning("Joint trajectory does not start at time 0.0, this is not officially supported.")

    n_frames = int(max(duration * fps, 1))
    frame_times = np.linspace(0.0, duration, n_frames)

    meshcat.StartRecording(set_visualizations_while_recording=False, frames_per_second=fps)

    for t in frame_times:
        context.SetTime(t)
        plant.SetPositions(plant_context, arm_index, joint_trajectory.value(t).squeeze())
        robot_diagram.ForcedPublish(context)

    meshcat.StopRecording()
    meshcat.PublishRecording()


def animate_joint_configurations(
    meshcat: Meshcat,
    robot_diagram: RobotDiagram,
    arm_index: ModelInstanceIndex,
    joint_configurations: list[np.ndarray] | JointPathType,
    duration: float = 4.0,
    time_per_configuration: float | None = None,
    context: Context | None = None,
) -> None:
    """Publish a recorded animation to meshcat where the robot arm cycles through the provided joint configurations.

    Args:
        meshcat: The MeshCat instance to add the visualization to.
        robot_diagram: The robot diagram.
        arm_index: The index of the robot arm to animate.
        joint_cofigurations: A list of joint configurations to animate.
        duration: Total duration of the animation.
        time_per_configuration: The time to spend (still) on each joint solution.  Will overwrite duration if provided.
        context: The context to use for the animation. If None, a new default context will be used (with all other joint angles equal to 0).
    """
    if joint_configurations is None or len(joint_configurations) == 0:
        logger.warning("Asked to animate joint configurations, but none provided.")
        return

    last_configuration = joint_configurations[-1].squeeze()
    joint_configurations = np.array(joint_configurations).squeeze()

    if time_per_configuration is not None:
        duration = time_per_configuration * len(joint_configurations)

    # Due to how ZeroOrderHold works, we need to duplicate the last configuration.
    # See note on how the second point is ignore when two points are provided:
    # https://drake.mit.edu/doxygen_cxx/classdrake_1_1trajectories_1_1_piecewise_polynomial.html#a06278ca4588ba1adb004b84fdfc62f3c
    joint_configurations = np.vstack([joint_configurations, last_configuration])

    joint_times = np.linspace(0.0, duration, len(joint_configurations))
    joints_interpolated = PiecewisePolynomial.ZeroOrderHold(joint_times, joint_configurations.T)

    animate_joint_trajectory(meshcat, robot_diagram, arm_index, joints_interpolated, context)


def animate_dual_joint_trajectory(
    meshcat: Meshcat,
    robot_diagram: RobotDiagram,
    arm_index_left: ModelInstanceIndex,
    arm_index_right: ModelInstanceIndex,
    joint_trajectory: DualArmTrajectory | Trajectory,
    degrees_of_freedom_left: int = 6,
    context: Context | None = None,
) -> None:
    """Publish a recorded animation to meshcat where the two robot arms follow the provided joint trajectories.

    Args:
        meshcat: The MeshCat instance to add the visualization to.
        robot_diagram: The robot diagram.
        arm_index_left: The index of the left robot arm to animate.
        arm_index_right: The index of the right robot arm to animate.
        joint_trajectory: The joint trajectory to animate.
        degrees_of_freedom_left: The number of degrees of freedom in the left arm, only used if joint_trajectory is a Drake Trajectory.
        context: The context to use for the animation. If None, a new default context will be used (with all other joint angles equal to 0).
    """
    if joint_trajectory is None:
        logger.warning("Asked to animate joint trajectory, but none provided.")  # can happen if e.g. planning fails

    is_stacked_drake_trajectory = False
    if isinstance(joint_trajectory, Trajectory):
        # Drake Trajectory, with all joint angles stacked
        is_stacked_drake_trajectory = True
        duration = joint_trajectory.end_time()
    else:
        positions_left = joint_trajectory.path_left.positions
        positions_right = joint_trajectory.path_right.positions
        times = joint_trajectory.times

        if positions_left is None or positions_right is None:
            logger.warning(
                "Asked to animate dual joint trajectory, where one of the path's positions is None is not supported yet."
            )
            return

        joint_trajectory_left = PiecewisePolynomial.FirstOrderHold(times, positions_left.T)
        joint_trajectory_right = PiecewisePolynomial.FirstOrderHold(times, positions_right.T)
        duration = joint_trajectory_left.end_time()

    if context is None:
        # Create a new context just for the recording
        context = robot_diagram.CreateDefaultContext()
    plant = robot_diagram.plant()
    plant_context = plant.GetMyContextFromRoot(context)

    fps = 60.0

    n_frames = int(max(duration * fps, 1))
    frame_times = np.linspace(0.0, duration, n_frames)

    meshcat.StartRecording(set_visualizations_while_recording=False, frames_per_second=fps)

    for t in frame_times:
        context.SetTime(t)
        if is_stacked_drake_trajectory:
            assert isinstance(joint_trajectory, Trajectory)  # For mypy
            plant.SetPositions(
                plant_context, arm_index_left, joint_trajectory.value(t)[:degrees_of_freedom_left].squeeze()
            )
            plant.SetPositions(
                plant_context, arm_index_right, joint_trajectory.value(t)[degrees_of_freedom_left:].squeeze()
            )
        else:
            plant.SetPositions(plant_context, arm_index_left, joint_trajectory_left.value(t).squeeze())
            plant.SetPositions(plant_context, arm_index_right, joint_trajectory_right.value(t).squeeze())
        robot_diagram.ForcedPublish(context)

    meshcat.StopRecording()
    meshcat.PublishRecording()



import airo_models
import numpy as np
from airo_typing import Vector3DType
from pydrake.math import RigidTransform
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.planning import RobotDiagramBuilder


def add_wall(
    robot_diagram_builder: RobotDiagramBuilder,
    type: str = "XZ",
    height: float = 2.0,
    width: float = 2.5,
    thickness: float = 0.2,
    position: Vector3DType = np.zeros(3),
    name: str = "wall",
) -> ModelInstanceIndex:
    """Adds a wall to the robot diagram builder.

    The wall is represented as a box with the specified dimensions, type, and position.
    The wall is automatically welded to the world frame for a static setup.

    Args:
        robot_diagram_builder: The robot diagram builder to which the wall will be added.
        type: The orientation of the wall.  Must be 'XZ' (parallel to the X-Z plane)
              or 'YZ' (parallel to the Y-Z plane).
        height: The height of the wall.
        width: The width of the wall (relevant for its orientation).
        thickness: The thickness of the wall (could be relevant for collision checking).
        position: The 3D position (x, y, z) of the center of the wall.
        name: The name of the wall (for easy reference).

    Returns:
        The ModelInstanceIndex of the added wall.
    """
    plant = robot_diagram_builder.plant()
    parser = robot_diagram_builder.parser()
    parser.SetAutoRenaming(True)

    z_size = height

    if type == "XZ":
        x_size = width
        y_size = thickness
    elif type == "YZ":
        x_size = thickness
        y_size = width
    else:
        raise ValueError("Invalid wall type, must be 'XZ' or 'YZ'")

    wall_urdf_path = airo_models.box_urdf_path((x_size, y_size, z_size), name)
    wall_index = parser.AddModels(wall_urdf_path)[0]

    wall_transform = RigidTransform(p=position)
    world_frame = plant.world_frame()
    wall_frame = plant.GetFrameByName("base_link", wall_index)

    plant.WeldFrames(world_frame, wall_frame, wall_transform)

    return wall_index



from pydrake.planning import RobotDiagramBuilder

from airo_drake import SingleArmScene, add_floor, add_manipulator, finish_build


def test_single_arm_scene():
    robot_diagram_builder = RobotDiagramBuilder()

    arm_index, gripper_index = add_manipulator(robot_diagram_builder, "ur5e", "robotiq_2f_85")
    add_floor(robot_diagram_builder)

    robot_diagram, context = finish_build(robot_diagram_builder)

    scene = SingleArmScene(robot_diagram, arm_index, gripper_index)

    del robot_diagram_builder

    assert robot_diagram is not None
    assert context is not None
    assert scene is not None
    assert arm_index is not None
    assert gripper_index is not None



import numpy as np
from airo_typing import JointPathType
from pydrake.multibody.optimization import CalcGridPointsOptions, Toppra
from pydrake.multibody.plant import MultibodyPlant
from pydrake.trajectories import PathParameterizedTrajectory, PiecewisePolynomial, Trajectory


def time_parametrize_toppra(
    plant: MultibodyPlant,
    joints: JointPathType | Trajectory,
    joint_speed_limit: float = 2.0,  # Max 180 degrees/s ~ 3.14 rad/s
    joint_acceleration_limit: float = 4.0,  # UR recommends < 800 degrees/s^2 ~ 13.9 rad/s^2
) -> PathParameterizedTrajectory:
    """Recalculate the timing of a path or trajectory to respect joint speed and acceleration limits using TOPP-RA.

    Args:
        plant: The MultibodyPlant for the robot.
        joints: A joint path or trajectory.
        joint_speed_limit: The maximum joint speed in rad/s.
        joint_acceleration_limit: The maximum joint acceleration in rad/s^2.
    """
    if isinstance(joints, Trajectory):
        joint_trajectory = joints
        n_dofs = joint_trajectory.value(0).shape[0]
    else:
        joints = np.array(joints).squeeze()
        n_dofs = joints.shape[1]
        times_dummy = np.linspace(0.0, 1.0, len(joints))  # TOPP-RA will calculate the actual times
        joint_trajectory = PiecewisePolynomial.FirstOrderHold(times_dummy, joints.T)

    gridpoints = Toppra.CalcGridPoints(joint_trajectory, CalcGridPointsOptions())

    acceleration_limits_lower = np.array([-joint_acceleration_limit] * n_dofs)
    acceleration_limits_upper = np.array([joint_acceleration_limit] * n_dofs)
    velocity_limits_lower = np.array([-joint_speed_limit] * n_dofs)
    velocity_limits_upper = np.array([joint_speed_limit] * n_dofs)

    toppra = Toppra(joint_trajectory, plant, gridpoints)
    toppra.AddJointAccelerationLimit(acceleration_limits_lower, acceleration_limits_upper)
    toppra.AddJointVelocityLimit(velocity_limits_lower, velocity_limits_upper)
    time_parametrization = toppra.SolvePathParameterization()

    if time_parametrization is None:
        raise ValueError("TOPP-RA failed to find a valid time parametrization.")

    joint_trajectory = PathParameterizedTrajectory(joint_trajectory, time_parametrization)

    return joint_trajectory



import numpy as np
from airo_typing import HomogeneousMatrixType
from pydrake.geometry import Cylinder, Meshcat, Rgba
from pydrake.math import RigidTransform, RotationMatrix


def visualize_frame(
    meshcat: Meshcat,
    name: str,
    transform: RigidTransform | HomogeneousMatrixType,
    length: float = 0.1,
    radius: float = 0.002,
    opacity: float = 1.0,
) -> None:
    """Visualizes a static frame in meshcat as an RGB (for xyz) colored triad.

    Args:
        transform: The 6D pose / transform of the frame to visualize.
        name: A name or path for the frame in the MeshCat hierarchy, can be used to delete it later.
        meshcat: The MeshCat instance to add the visualization to.
        length: Length of each axis in the triad.
        radius: Radius of each axis cylinder.
        opacity: Opacity for the axis colors.
    """

    meshcat.SetTransform(name, transform)
    axis_labels = ["x-axis", "y-axis", "z-axis"]
    axis_colors = [
        Rgba(1, 0, 0, opacity),
        Rgba(0, 1, 0, opacity),
        Rgba(0, 0, 1, opacity),
    ]
    axis_transforms = [
        RigidTransform(RotationMatrix.MakeYRotation(np.pi / 2), [length / 2.0, 0, 0]),  # type: ignore
        RigidTransform(RotationMatrix.MakeXRotation(np.pi / 2), [0, length / 2.0, 0]),  # type: ignore
        RigidTransform([0, 0, length / 2.0]),  # type: ignore
    ]

    for transform, color, label in zip(axis_transforms, axis_colors, axis_labels):
        meshcat.SetTransform(f"{name}/{label}", transform)
        meshcat.SetObject(f"{name}/{label}", Cylinder(radius, length), color)



from pydrake.trajectories import PathParameterizedTrajectory, PiecewisePolynomial, Trajectory


def shift_drake_trajectory_in_time(trajectory: Trajectory, time_shift: float) -> PathParameterizedTrajectory:
    """Shifts a Drake trajectory in time by a specified amount.

    Args:
        trajectory: The Drake trajectory to be shifted.
        time_shift: The amount of time (in seconds) to shift the trajectory forward.

    Returns:
        A new PathParameterizedTrajectory representing the shifted trajectory.
    """
    start = trajectory.start_time()
    end = trajectory.end_time()
    time_shift_trajectory = PiecewisePolynomial.FirstOrderHold([start + time_shift, end + time_shift], [[start, end]])
    shifted_trajectory = PathParameterizedTrajectory(trajectory, time_shift_trajectory)

    return shifted_trajectory



import itertools
from typing import List, Tuple

from airo_typing import JointConfigurationType
from pydrake.all import BodyIndex
from pydrake.planning import RobotCollisionType, SceneGraphCollisionChecker


def filter_collisions_between_all_body_pairs(
    collision_checker: SceneGraphCollisionChecker,
    body_indices_one: List[BodyIndex],
    body_indices_two: List[BodyIndex],
    filtered: bool = True,
) -> None:
    """Enable or disable collision filtering between all pairs of the bodies listed in body_indices_one and body_indices_two.
    When collision filtering is enabled for two bodies, they are allowed to collide.

    Args:
        collision_checker: The collision checker instance to alter.
        body_indices_one: A list of body indices.
        body_indices_two: A list of body indices.
        filtered: Whether to filter collisions between these bodies."""

    body_combinations = itertools.product(body_indices_one, body_indices_two)
    for body_index_1, body_index_2 in body_combinations:
        collision_checker.SetCollisionFilteredBetween(body_index_1, body_index_2, filtered)


def list_collisions_between_bodies(
    collision_checker: SceneGraphCollisionChecker, q: JointConfigurationType
) -> List[Tuple[BodyIndex, BodyIndex, bool]]:
    """List all collisions between bodies for a joint configuration.
    The return type of this function is a list that contains all colliding bodies (if any) and a boolean value that indicates whether the collision is a self-collision.
    When the boolean value is false, the collision is an environment collision.

    Args:
        collision_checker: The collision checker instance to use.
        q: The joint configuration of the robot to check.

    Returns:
        A list of tuples of colliding body indices and a boolean value that is True when the collision is a self collision."""
    robot_clearance = collision_checker.CalcRobotClearance(q, 1.0)
    indices_one = robot_clearance.robot_indices()
    indices_two = robot_clearance.other_indices()
    collision_types = robot_clearance.collision_types()
    distances = robot_clearance.distances()

    return [
        (tup[0], tup[1], tup[2] != RobotCollisionType.kEnvironmentCollision)
        for tup in zip(indices_one, indices_two, collision_types, distances)
        if tup[3] <= 0.0
    ]



# init files are still required in modern python!
# https://peps.python.org/pep-0420/ introduced implicit namespace packages
# but for building and many toolings, you still need to have __init__ files (at least in the root of the package).
# e.g. if you remove this init file and try to build with pip install .
# you won't be able to import the dummy module.


# We disable isort for the entire __init__.py file, because the order of imports is important here.
# If a function depends on another function, the dependent function should be imported after the function it depends on.
# isort: skip_file
from airo_drake.building.finish import finish_build
from airo_drake.building.floor import add_floor
from airo_drake.building.manipulator import X_URBASE_ROSBASE, X_URTOOL0_ROBOTIQ, add_manipulator
from airo_drake.building.meshcat import add_meshcat
from airo_drake.building.wall import add_wall
from airo_drake.path.analysis import (
    calculate_joint_path_distances,
    calculate_joint_path_length,
    calculate_joint_path_outlier_threshold,
    find_closest_configuration,
    joint_path_has_large_jumps,
)
from airo_drake.path.processing import (
    calculate_joint_paths,
    calculate_path_ik_solutions,
    calculate_valid_joint_paths,
    create_paths_from_closest_solutions,
)
from airo_drake.scene import DualArmScene, SingleArmScene
from airo_drake.trajectory.timing import shift_drake_trajectory_in_time
from airo_drake.trajectory.concatenate import concatenate_drake_trajectories
from airo_drake.time_parametrization.toppra import time_parametrize_toppra
from airo_drake.trajectory.discretize import discretize_drake_joint_trajectory, discretize_drake_pose_trajectory
from airo_drake.trajectory.interpolate import joint_trajectory_to_drake
from airo_drake.visualization.frame import visualize_frame
from airo_drake.visualization.joints import (
    animate_dual_joint_trajectory,
    animate_joint_configurations,
    animate_joint_trajectory,
)
from airo_drake.collision.collision_checking import (
    filter_collisions_between_all_body_pairs,
    list_collisions_between_bodies,
)


__all__ = [
    "add_floor",
    "add_wall",
    "add_manipulator",
    "X_URBASE_ROSBASE",
    "X_URTOOL0_ROBOTIQ",
    "add_meshcat",
    "finish_build",
    "SingleArmScene",
    "DualArmScene",
    "visualize_frame",
    "animate_joint_configurations",
    "animate_joint_trajectory",
    "animate_dual_joint_trajectory",
    "time_parametrize_toppra",
    "discretize_drake_joint_trajectory",
    "discretize_drake_pose_trajectory",
    "joint_trajectory_to_drake",
    "shift_drake_trajectory_in_time",
    "concatenate_drake_trajectories",
    "calculate_joint_path_outlier_threshold",
    "calculate_joint_path_length",
    "joint_path_has_large_jumps",
    "find_closest_configuration",
    "calculate_joint_path_distances",
    "create_paths_from_closest_solutions",
    "calculate_path_ik_solutions",
    "calculate_joint_paths",
    "calculate_valid_joint_paths",
    "filter_collisions_between_all_body_pairs",
    "list_collisions_between_bodies",
]



import airo_models
import numpy as np
from airo_typing import HomogeneousMatrixType
from pydrake.math import RigidTransform, RollPitchYaw
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.planning import RobotDiagramBuilder

# X_URBASE_ROSBASE is the 180 rotation between ROS URDF base and the UR control box base
X_URBASE_ROSBASE = RigidTransform(rpy=RollPitchYaw([0, 0, np.pi]), p=[0, 0, 0])  # type: ignore

# 180 degree rotation and 1 cm (estimate) offset for the coupling / flange
X_URTOOL0_ROBOTIQ = RigidTransform(rpy=RollPitchYaw([0, 0, np.pi / 2]), p=[0, 0, 0.01])  # type: ignore


def add_manipulator(
    robot_diagram_builder: RobotDiagramBuilder,
    name: str,
    gripper_name: str,
    arm_transform: HomogeneousMatrixType | None = None,
    gripper_transform: HomogeneousMatrixType | None = None,
    static_gripper: bool = False,
) -> tuple[ModelInstanceIndex, ModelInstanceIndex]:
    """Add a manipulator (a robot arm with a gripper) to the robot diagram builder.
    Looks up the URDF files for the robot and gripper and welds them together.
    Also provides slightly opionatated default transforms for the welds.
    For example, we rotate the ROS UR URDFs 180 degrees. This enables us to send
    TCP poses in the Drake world frame to the UR control box.

    Args:
        robot_diagram_builder: The robot diagram builder to which the manipulator will be added.
        name: The name of the robot arm, must be known by airo-models
        gripper_name: The name of the gripper, must be known by airo-models
        arm_transform: The transform of the robot arm, if None, we use supply a robot-specific default.
        gripper_transform: The transform of the gripper, if None, we supply a default for the robot-gripper pair.
        static_gripper: If True, will fix all gripper joints to their default. Useful when you don't want the gripper DoFs in the plant.

    Returns:
        The robot and gripper index.
    """

    plant = robot_diagram_builder.plant()
    parser = robot_diagram_builder.parser()
    parser.SetAutoRenaming(True)

    # Load URDF files
    arm_urdf_path = airo_models.get_urdf_path(name)
    gripper_urdf_path = airo_models.get_urdf_path(gripper_name)

    arm_index = parser.AddModels(arm_urdf_path)[0]

    if static_gripper:
        gripper_urdf = airo_models.urdf.read_urdf(gripper_urdf_path)
        airo_models.urdf.make_static(gripper_urdf)
        gripper_urdf_path = airo_models.urdf.write_urdf_to_tempfile(
            gripper_urdf, gripper_urdf_path, prefix=f"{gripper_name}_static_"
        )

    gripper_index = parser.AddModels(gripper_urdf_path)[0]

    # Weld some frames together
    world_frame = plant.world_frame()
    arm_frame = plant.GetFrameByName("base_link", arm_index)
    arm_tool_frame = plant.GetFrameByName("tool0", arm_index)
    gripper_frame = plant.GetFrameByName("base_link", gripper_index)

    if arm_transform is None:
        if name.startswith("ur"):
            arm_rigid_transform = X_URBASE_ROSBASE
        else:
            arm_rigid_transform = RigidTransform()
    else:
        arm_rigid_transform = RigidTransform(arm_transform)

    if gripper_transform is None:
        if name.startswith("ur") and gripper_name.startswith("robotiq"):
            gripper_rigid_transform = X_URTOOL0_ROBOTIQ
        else:
            gripper_rigid_transform = RigidTransform()
    else:
        gripper_rigid_transform = RigidTransform(gripper_transform)

    plant.WeldFrames(world_frame, arm_frame, arm_rigid_transform)
    plant.WeldFrames(arm_tool_frame, gripper_frame, gripper_rigid_transform)

    return arm_index, gripper_index



from dataclasses import dataclass

from pydrake.geometry import Meshcat
from pydrake.multibody.tree import ModelInstanceIndex
from pydrake.planning import RobotDiagram


@dataclass
class SingleArmScene:
    """The most important objects when using Drake with a single robot arm."""

    robot_diagram: RobotDiagram
    arm_index: ModelInstanceIndex
    gripper_index: ModelInstanceIndex | None = None
    meshcat: Meshcat | None = None


@dataclass
class DualArmScene:
    """The most important objects when using Drake with two robot arms."""

    robot_diagram: RobotDiagram
    arm_left_index: ModelInstanceIndex
    arm_right_index: ModelInstanceIndex
    gripper_left_index: ModelInstanceIndex | None = None
    gripper_right_index: ModelInstanceIndex | None = None
    meshcat: Meshcat | None = None



from pydrake.geometry import Meshcat
from pydrake.planning import RobotDiagram, RobotDiagramBuilder
from pydrake.systems.framework import Context
from pydrake.visualization import ApplyVisualizationConfig, VisualizationConfig


def finish_build(
    robot_diagram_builder: RobotDiagramBuilder, meshcat: Meshcat | None = None
) -> tuple[RobotDiagram, Context]:
    """Finish building the robot diagram and create a context. The reason we create and force publish a context it so
    that you can see your robot models visualized in Meshcat. Also if a meshcat is provided, we enable some extra
    visualization such as contacts. We do this here because this can only be done after the plant is finalized.

    Note that after finishing the build, we can no longer add new objects to the Drake scene. However this needs to be
    done to be able to use many functionalities, e.g. collision checking. This is the standard workflow in Drake and a
    known "limitation".

    Args:
        robot_diagram_builder: The RobotDiagramBuilder object to which all models have already been added.
        meshcat: The Meshcat object.

    Returns:
        diagram: The robot diagram.
        context: A default context that you can use as you wish.
    """
    plant = robot_diagram_builder.plant()

    if meshcat is not None:
        builder = robot_diagram_builder.builder()
        plant.Finalize()
        config = VisualizationConfig(publish_contacts=True, enable_alpha_sliders=True)
        ApplyVisualizationConfig(config, builder=builder, plant=plant, meshcat=meshcat)

    robot_diagram = robot_diagram_builder.Build()

    context = robot_diagram.CreateDefaultContext()
    robot_diagram.ForcedPublish(context)

    return robot_diagram, context



import numpy as np
from airo_typing import JointPathContainer, PoseTrajectory, SingleArmTrajectory
from pydrake.trajectories import Trajectory


def discretize_drake_joint_trajectory(joint_trajectory: Trajectory, steps: int = 100) -> SingleArmTrajectory:
    positions = []
    times_uniform = np.linspace(joint_trajectory.start_time(), joint_trajectory.end_time(), steps)
    for t in times_uniform:
        position = joint_trajectory.value(t).squeeze()
        positions.append(position)
    joint_path = JointPathContainer(positions=np.array(positions))
    joint_trajectory_discretized = SingleArmTrajectory(times_uniform, joint_path)
    return joint_trajectory_discretized


def discretize_drake_pose_trajectory(pose_trajectory: Trajectory, steps: int = 100) -> PoseTrajectory:
    poses = []
    times_uniform = np.linspace(pose_trajectory.start_time(), pose_trajectory.end_time(), steps)
    for t in times_uniform:
        pose = pose_trajectory.value(t).squeeze()
        poses.append(np.array(pose))

    poses_array = np.array(poses)
    pose_trajectory_discretized = PoseTrajectory(times_uniform, poses_array)
    return pose_trajectory_discretized



from airo_typing import SingleArmTrajectory
from pydrake.trajectories import PiecewisePolynomial


def joint_trajectory_to_drake(joint_trajectory: SingleArmTrajectory) -> PiecewisePolynomial:
    """Converts an airo-mono SingleArmTrajectory to a Drake PiecewisePolynomial trajectory, using linear interpolation between configurations.

    Args:
        joint_trajectory: The discretized trajectory to be converted (and thus interpolated).

    Returns:
        A Drake trajectory representing the equivalent joint motion,
            using first-order hold (linear interpolation).

    Raises:
        ValueError: If the input SingleArmTrajectory does not contain position data.
    """
    positions = joint_trajectory.path.positions
    times = joint_trajectory.times
    if positions is None:
        raise ValueError("joint_trajectory has no positions.")

    # FirstOrderHold is linear interpolation
    joint_trajectory_interpolated = PiecewisePolynomial.FirstOrderHold(times, positions.T)

    return joint_trajectory_interpolated



import numpy as np
from airo_typing import JointConfigurationType, JointPathType


def find_closest_configuration(
    reference_configuration: JointConfigurationType, candidates: list[JointConfigurationType]
) -> JointConfigurationType:
    """Finds the closest configuration to a reference configuration within a set of candidates."""
    candidates_array = np.array(candidates).squeeze()
    distances = np.linalg.norm(candidates_array - reference_configuration, axis=1)  # Vectorized calculation
    closest_index = np.argmin(distances)
    return candidates_array[closest_index]


def calculate_joint_path_distances(path: JointPathType) -> np.ndarray:
    """Calculate the distances between consecutive joint configurations in a joint path.

    Args:
        path: A path of joint configurations.

    Returns:
        An array of distances between consecutive joint configurations (1 shorter than path).
    """
    return np.linalg.norm(np.diff(path, axis=0), axis=1)


def calculate_joint_path_length(path: JointPathType) -> float:
    """Calculate the length of a joint path.

    Args:
        path: A path of joint configurations.

    Returns:
        The length of the joint path.
    """
    return np.sum(calculate_joint_path_distances(path))


def calculate_joint_path_outlier_threshold(distances: np.ndarray, iqr_multiplier: float = 20.0) -> float:
    """Calculate a threshold for detecting unusually large jumps in a joint path.

    We use a one-sided IQR-based (interquartile range) method:
    * One-sided because we don't care about unusually small jumps.
    * IQR-based instead of using standard deviation because it's less sensitive to outliers.

    Jumps between consecutive IK solutions can happen for a few reasons:
    * Singularities: Near robot singularities, small changes in the end-effector
    pose can lead to large, discontinuous changes in the joint configurations required.
    * Incompleteness: the IK solver might not find the solution that is closest to the previous one.

    Args:
        distance: The distances between consecutive joint configurations.
        iqr_multiplier: A multiplier for the interquartile range.

    Returns:
        A threshold for detecting unusually large jumps in joint configuration distances.
    """
    q1, q3 = np.percentile(distances, [25, 75])
    interquartile_range = q3 - q1
    threshold = q3 + (iqr_multiplier * interquartile_range)
    return threshold


def joint_path_has_large_jumps(path: JointPathType, iqr_multiplier: float = 20.0) -> bool:
    """Check a joint path for unusually large jumps in joint configuration distances.

    See the docstring for `calculate_joint_path_iqr_threshold` for more details.

    Args:
        path: A path of joint configurations.
        iqr_multiplier: A multiplier for the interquartile range.

    Returns:
        A boolean indicating whether the joint path contains unusually large jumps.
    """
    distances = calculate_joint_path_distances(path)
    threshold = calculate_joint_path_outlier_threshold(distances, iqr_multiplier)
    return bool(np.any(distances > threshold))
